import json
import os
import re
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Dict

import boto3

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.aws import get_profile_name_from_account_id
from dbt_platform_helper.utils.aws import get_ssm_secrets
from dbt_platform_helper.utils.messages import abort_with_error


@dataclass
class Environment:
    name: str
    account_id: str
    sessions: Dict[str, boto3.Session]

    @property
    def session(self):
        if self.account_id not in self.sessions:
            self.sessions[self.account_id] = get_aws_session_or_abort(
                get_profile_name_from_account_id(self.account_id),
            )

        return self.sessions[self.account_id]


@dataclass
class Service:
    name: str
    kind: str


@dataclass
class Application:
    name: str
    environments: Dict[str, Environment] = field(default_factory=dict)
    services: Dict[str, Service] = field(default_factory=dict)

    def __str__(self):
        output = f"Application {self.name} with"

        environments = [f"{env.name}:{env.account_id}" for env in self.environments.values()]

        if environments:
            return f"{output} environments {', '.join(environments)}"

        return f"{output} no environments"

    def __eq__(self, other):
        return str(self) == str(other)


def load_application(app=None, default_session=None, env=None) -> Application:
    application = Application(app if app else get_application_name())
    current_session = default_session if default_session else get_aws_session_or_abort()

    ssm_client = current_session.client("ssm")
    sts_client = current_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    sessions = {account_id: current_session}

    def is_environment_key(name):
        """
        Match only parameter names that are an environment path with no further
        nesting.

        e.g.
         - /platform/applications/test/environments/my_env will match.
         - /copilot/applications/test/environments/my_env will match.
         - /copilot/applications/test/environments/my_env/addons will not match.
        """
        environment_key_regex = r"^/(copilot|platform)/applications/{}/environments/[^/]*$".format(
            application.name
        )
        return bool(re.match(environment_key_regex, name))

    environments_data = []

    # Try to load all /platform SSM parameters that are present
    env_params = get_ssm_secrets(
        app=app,
        env=None,
        session=current_session,
        path=f"/platform/applications/{application.name}/environments",
    )

    if env_params:
        for name, value in env_params:
            try:
                param_data = json.loads(value)
            except json.JSONDecodeError:
                continue

            # Each /platform SSM parameter contains data about all the environments of an application
            if "allEnvironments" in param_data:
                environments_data = param_data["allEnvironments"]
                break  # Only need one
    else:
        try:
            # Check that the Copilot application exists
            ssm_client.get_parameter(
                Name=f"/copilot/applications/{application.name}",
                WithDecryption=False,
            )

            # Legacy /copilot SSM parameters for each environment
            env_params = get_ssm_secrets(
                app, None, current_session, f"/copilot/applications/{application.name}/environments"
            )

            for name, value in env_params:
                try:
                    param_data = json.loads(value)
                except json.JSONDecodeError:
                    continue

                if is_environment_key(name):
                    environments_data.append(param_data)

        except ssm_client.exceptions.ParameterNotFound:
            raise ApplicationNotFoundException(
                application_name=application.name, environment_name=env
            )

    application.environments = {
        env["name"]: Environment(env["name"], env["accountID"], sessions)
        for env in environments_data
    }

    application.services = _load_services(ssm_client, application)

    return application


def _load_services(ssm_client, application: Application) -> Dict[str, Service]:
    """
    Try to load
    /platform/applications/{app}/environments/{env}/services/{service}
    parameters if present.

    Otherwise, fall back to legacy /copilot/applications/{app}/components
    parameters.
    """
    services: Dict[str, Service] = {}

    # Try /platform SSM parameter
    for env_name in application.environments.keys():
        params = dict(
            Path=f"/platform/applications/{application.name}/environments/{env_name}/services",
            Recursive=False,
            WithDecryption=False,
        )

        while True:
            response = ssm_client.get_parameters_by_path(**params)
            for ssm_param in response.get("Parameters", []):
                try:
                    data = json.loads(ssm_param["Value"])
                    name = data["name"]
                    kind = data["type"]
                    services.setdefault(name, Service(name, kind))  # Avoid duplicates
                except (json.JSONDecodeError, KeyError):
                    continue

            if "NextToken" in response:
                params["NextToken"] = response["NextToken"]
            else:
                break

    if services:
        return services

    # Fallback to legacy /copilot SSM parameter
    response = ssm_client.get_parameters_by_path(
        Path=f"/copilot/applications/{application.name}/components",
        Recursive=False,
        WithDecryption=False,
    )
    results = response["Parameters"]
    while "NextToken" in response:
        response = ssm_client.get_parameters_by_path(
            Path=f"/copilot/applications/{application.name}/components",
            Recursive=False,
            WithDecryption=False,
            NextToken=response["NextToken"],
        )
        results.extend(response["Parameters"])

    legacy_services = {
        svc["name"]: Service(svc["name"], svc["type"])
        for svc in [json.loads(parameter["Value"]) for parameter in results]
    }

    return legacy_services


def get_application_name(abort=abort_with_error):
    if Path(PLATFORM_CONFIG_FILE).exists():
        config = ConfigProvider(installed_version_provider="N/A")
        try:
            app_config = config.load_unvalidated_config_file()
            return app_config["application"]
        except KeyError:
            abort(
                f"Cannot get application name. No 'application' key can be found in {PLATFORM_CONFIG_FILE}"
            )
    else:
        abort(f"Cannot get application name. {PLATFORM_CONFIG_FILE} is missing.")


class ApplicationException(PlatformException):
    pass


class ApplicationNotFoundException(ApplicationException):
    def __init__(self, application_name: str, environment_name: str):
        super().__init__(
            f"""The account "{os.environ.get("AWS_PROFILE")}" does not contain the application "{application_name}". 
Please ensure that the environment variable "AWS_PROFILE" is set correctly. If the issue persists, verify that one of the following AWS SSM parameters exists:
 - /platform/applications/{application_name}/environments/{environment_name}
 - /copilot/applications/{application_name}"""
        )


class ApplicationServiceNotFoundException(ApplicationException):
    def __init__(self, application_name: str, svc_name: str):
        super().__init__(
            f"""The service {svc_name} was not found in the application {application_name}. It either does not exist, or has not been deployed."""
        )


class ApplicationEnvironmentNotFoundException(ApplicationException):
    def __init__(self, application_name: str, environment: str):
        super().__init__(
            f"""The environment "{environment}" either does not exist or has not been deployed for the application {application_name}."""
        )



import re
from enum import Enum
from typing import ClassVar
from typing import Dict
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator

from dbt_platform_helper.platform_exception import PlatformException


class HealthCheck(BaseModel):
    path: Optional[str] = Field(
        description="The destination that the health check requests are sent to.", default="/"
    )
    port: Optional[int] = Field(
        description="The port that the health check requests are sent to.", default=8080
    )
    success_codes: Optional[str] = Field(
        description="A comma-separated list of HTTP status codes that healthy targets must use when responding to a HTTP health check.",
        default="200",
    )
    healthy_threshold: Optional[int] = Field(
        description="The number of consecutive health check successes required before considering an unhealthy target healthy.",
        default=3,
    )
    unhealthy_threshold: Optional[int] = Field(
        description="The number of consecutive health check failures required before considering a target unhealthy.",
        default=3,
    )
    interval: Optional[int] = Field(
        description="The approximate amount of time, in seconds, between health checks of an individual target.",
        default=35,
    )
    timeout: Optional[int] = Field(
        description="The amount of time, in seconds, during which no response from a target means a failed health check.",
        default=30,
    )
    grace_period: Optional[int] = Field(
        description="The amount of time to ignore failing target group healthchecks on container start.",
        default=30,
    )


class AdditionalRules(BaseModel):
    path: str = Field(description="""Requests to this path will be forwarded to your service.""")
    alias: list[str] = Field(description="""The HTTP domain alias of the service.""")


class Http(BaseModel):
    alias: list[str] = Field(
        description="List of HTTPS domain alias(es) of your service.", default=None
    )
    stickiness: Optional[bool] = Field(description="Enable sticky sessions.", default=False)
    path: str = Field(description="Requests to this path will be forwarded to your service.")
    target_container: str = Field(description="Target container for the requests.")
    healthcheck: HealthCheck = Field(default_factory=HealthCheck)
    additional_rules: Optional[list[AdditionalRules]] = Field(default=None)
    deregistration_delay: Optional[int] = Field(
        default=60,
        description="The amount of time to wait for targets to drain connections during deregistration.",
    )


class HttpOverride(BaseModel):
    alias: Optional[list[str]] = Field(
        description="List of HTTPS domain alias(es) of your service.", default=None
    )
    stickiness: Optional[bool] = Field(description="Enable sticky sessions.", default=None)
    path: Optional[str] = Field(
        description="Requests to this path will be forwarded to your service.", default=None
    )
    target_container: Optional[str] = Field(
        description="Target container for the requests", default=None
    )
    healthcheck: Optional[HealthCheck] = Field(default=None)
    additional_rules: Optional[list[AdditionalRules]] = Field(default=None)
    deregistration_delay: Optional[int] = Field(
        default=None,
        description="The amount of time to wait for targets to drain connections during deregistration.",
    )


class ContainerHealthCheck(BaseModel):
    command: list[str] = Field(
        description="The command to run to determine if the container is healthy."
    )
    interval: Optional[int] = Field(
        default=10, description="Time period between health checks, in seconds."
    )
    retries: Optional[int] = Field(
        default=2, description="Number of times to retry before container is deemed unhealthy."
    )
    timeout: Optional[int] = Field(
        default=5,
        description="How long to wait before considering the health check failed, in seconds.",
    )
    start_period: Optional[int] = Field(
        default=0,
        description="Length of grace period for containers to bootstrap before failed health checks count towards the maximum number of retries.",
    )


class Sidecar(BaseModel):
    port: int = Field(description="Container port exposed by the sidecar to receive traffic.")
    image: str = Field(description="Container image URI for the sidecar (e.g. 'repo/image:tag').")
    essential: Optional[bool] = Field(
        description="Whether the ECS task should stop if this sidecar container exits.",
        default=True,
    )
    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(
        description="Environment variables to inject into the sidecar container.", default=None
    )
    secrets: Optional[Dict[str, str]] = Field(
        description="Parameter Store secrets to inject into the sidecar.", default=None
    )
    healthcheck: Optional[ContainerHealthCheck] = Field(default=None)


class SidecarOverride(BaseModel):
    port: Optional[int] = Field(default=None)
    image: Optional[str] = Field(default=None)
    essential: Optional[bool] = Field(default=None)
    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(default=None)
    secrets: Optional[Dict[str, str]] = Field(default=None)
    healthcheck: Optional[ContainerHealthCheck] = Field(default=None)


class Image(BaseModel):
    location: str = Field(description="Main container image location.")
    port: Optional[int] = Field(
        description="Port exposed by the main ECS task container (used by the load balancer/Service Connect).",
        default=None,
    )
    depends_on: Optional[dict[str, str]] = Field(
        description="Container dependency conditions.", default=None
    )
    healthcheck: Optional[ContainerHealthCheck] = Field(default=None)

    @field_validator("location", mode="after")
    @classmethod
    def is_image_untagged(cls, value: str) -> str:
        image_name = value.split("/")[-1]
        if ":" in image_name:
            raise PlatformException(
                f"Image location cannot contain a tag '{value}'\nPlease remove the tag from your image location. The image tag is automatically added during deployment."
            )
        return value


class Storage(BaseModel):
    readonly_fs: Optional[bool] = Field(
        description="Specify true to give your container read-only access to its root file system.",
        default=False,
    )
    writable_directories: Optional[list[str]] = Field(
        description="List of directories with read/write access.", default=None
    )
    ephemeral: Optional[int] = Field(
        description="The total amount, in GiB, of ephemeral storage to set for the ECS task.",
        default=None,
    )

    @field_validator("writable_directories", mode="after")
    @classmethod
    def has_leading_forward_slash(cls, value: Union[list, None]) -> Union[list, None]:
        if value is not None:
            for path in value:
                if not path.startswith("/"):
                    raise PlatformException(
                        "All writable directory paths must be absolute (starts with a /)"
                    )
        return value

    @model_validator(mode="after")
    def is_within_allowed_range(self):
        if self.ephemeral:
            if self.ephemeral < 21 or self.ephemeral > 200:
                raise PlatformException(
                    "The minimum supported ephemeral storage value is 21 (GiB) and the maximum supported value is 200 (GiB)."
                )
        return self


class Cooldown(BaseModel):
    in_: Optional[int] = Field(
        alias="in",
        description="Number of seconds to wait before scaling in (down) after a drop in load.",
        default=60,
    )  # Can't use 'in' because it's a reserved keyword
    out: Optional[int] = Field(
        description="Number of seconds to wait before scaling out (up) after a spike in load.",
        default=60,
    )

    @field_validator("in_", "out", mode="before")
    @classmethod
    def parse_seconds(cls, value):
        if isinstance(value, str) and value.endswith("s"):
            value = value.removesuffix("s")  # remove the trailing 's'
        try:
            return int(value)
        except (ValueError, TypeError):
            raise PlatformException("Cooldown values must be integers or strings like '30s'")


class CpuPercentage(BaseModel):
    value: int = Field(description="Target CPU utilisation percentage that triggers autoscaling.")
    cooldown: Optional[Cooldown] = Field(
        default=None, description="Optional CPU cooldown that overrides the global cooldown policy."
    )


class MemoryPercentage(BaseModel):
    value: int = Field(description="Target CPU utilisation percentage that triggers autoscaling.")
    cooldown: Optional[Cooldown] = Field(
        default=None,
        description="Optional memory cooldown that overrides the global cooldown policy.",
    )


class RequestsPerMinute(BaseModel):
    value: int = Field(
        description="Number of incoming requests per minute that triggers autoscaling."
    )
    cooldown: Optional[Cooldown] = Field(
        default=None,
        description="Optional requests cooldown that overrides the global cooldown policy.",
    )


class Count(BaseModel):
    range: str = Field(
        description="Minimum and maximum number of ECS tasks to maintain e.g. '1-2'."
    )
    cooldown: Optional[Cooldown] = Field(
        default=None,
        description="Global cooldown applied to all autoscaling metrics unless overridden per metric.",
    )
    cpu_percentage: Optional[Union[int, CpuPercentage]] = Field(
        default=None,
        description="CPU utilisation threshold (0–100). Either a plain integer or a map with 'value' and 'cooldown'.",
    )
    memory_percentage: Optional[Union[int, MemoryPercentage]] = Field(
        default=None,
        description="Memory utilisation threshold (0–100). Either a plain integer or a map with 'value' and 'cooldown'.",
    )
    requests_per_minute: Optional[Union[int, RequestsPerMinute]] = Field(
        default=None,
        description="Request-rate threshold. Either a plain integer or a map with 'value' and 'cooldown'.",
    )

    @model_validator(mode="after")
    def at_least_one_autoscaling_metric(self):

        if not any([self.cpu_percentage, self.memory_percentage, self.requests_per_minute]):
            raise PlatformException(
                "If autoscaling is enabled, you must define at least one metric: "
                "cpu_percentage, memory_percentage, or requests_per_minute"
            )

        if not re.match(r"^(\d+)-(\d+)$", self.range):
            raise PlatformException("Range must be in the format 'int-int' e.g. '1-2'")

        range_split = self.range.split("-")
        if int(range_split[0]) >= int(range_split[1]):
            raise PlatformException("Range minimum value must be less than the maximum value.")

        return self


class ServiceConfigEnvironmentOverride(BaseModel):
    http: Optional[HttpOverride] = Field(default=None)
    sidecars: Optional[Dict[str, SidecarOverride]] = Field(default=None)
    image: Optional[Image] = Field(default=None)

    cpu: Optional[int] = Field(default=None)
    memory: Optional[int] = Field(default=None)
    count: Optional[Union[int, Count]] = Field(default=None)
    exec: Optional[bool] = Field(default=None)
    entrypoint: Optional[list[str]] = Field(default=None)
    essential: Optional[bool] = Field(default=None)

    storage: Optional[Storage] = Field(default=None)

    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(default=None)
    secrets: Optional[Dict[str, str]] = Field(default=None)


class ServiceType(str, Enum):
    BACKEND_SERVICE = "Backend Service"
    LOAD_BALANCED_WEB_SERVICE = "Load Balanced Web Service"


class ServiceConfig(BaseModel):
    name: str = Field(description="Service name.")
    type: ServiceType = Field(
        description=f"Type of service. Must one one of: '{ServiceType.LOAD_BALANCED_WEB_SERVICE.value}', '{ServiceType.BACKEND_SERVICE.value}'"
    )
    http: Optional[Http] = Field(default=None)

    @model_validator(mode="after")
    def check_http_for_web_service(self):
        if self.type == ServiceType.LOAD_BALANCED_WEB_SERVICE and self.http is None:
            raise PlatformException(
                f"A 'http' block must be provided when service type == {self.type.value}"
            )
        return self

    sidecars: Optional[Dict[str, Sidecar]] = Field(default=None)
    image: Image = Field()
    cpu: int = Field(
        description="vCPU units reserved for the ECS task (e.g. 256=0.25 vCPU, 512=0.5 vCPU, 1024=1 vCPU)."
    )
    memory: int = Field(
        description="Memory in MiB reserved for the ECS task (e.g. 256, 512, 1024)."
    )
    count: Union[int, Count] = Field(
        description="Desired task count — either a fixed integer or an autoscaling policy map with 'range', 'cooldown', and at least one of 'cpu_percentage', 'memory_percentage', or 'requests_per_minute' metrics."
    )
    exec: Optional[bool] = Field(
        description="Enable ECS Exec (remote command execution) for running ECS tasks.",
        default=False,
    )
    entrypoint: Optional[list[str]] = Field(
        description="Overrides the default entrypoint in the image.", default=None
    )
    essential: Optional[bool] = Field(
        description="Whether the main container is marked essential; The entire ECS task stops if it exits.",
        default=True,
    )
    storage: Storage = Field(default_factory=Storage)
    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(
        description="Environment variables to inject into the main application container.",
        default=None,
    )
    secrets: Optional[Dict[str, str]] = Field(
        description="Parameter Store secrets to inject into the main application container.",
        default=None,
    )
    # Environment overrides can override almost the full config
    environments: Optional[Dict[str, ServiceConfigEnvironmentOverride]] = Field(
        description="Allows you to override most service config properties for specific environments.",
        default=None,
    )

    # Class based variable used when handling the object
    local_terraform_source: ClassVar[str] = "../../../../../platform-tools/terraform/ecs-service"



from click import Argument
from click import Choice
from click import Command
from click import Context
from click import Group
from click import HelpFormatter


class ClickDocOptCommand(Command):
    def format_usage(self, ctx: Context, formatter: HelpFormatter) -> None:
        format_click_usage(ctx, formatter)


class ClickDocOptGroup(Group):
    command_class = ClickDocOptCommand

    def format_usage(self, ctx: Context, formatter: HelpFormatter) -> None:
        format_click_usage(ctx, formatter, True)


def format_click_usage(ctx: Context, formatter: HelpFormatter, group: bool = False) -> None:
    help_text = f"Usage: {ctx.command_path} "
    current_line = 0
    indent = len(help_text)

    parameters = list(ctx.command.params)
    parameters.sort(key=lambda p: p.required, reverse=True)
    parameters.sort(key=lambda p: hasattr(p, "is_flag") and p.is_flag)
    parameters.sort(key=lambda p: p.__class__.__name__ == "Option")

    if group:
        command_list = list(ctx.command.commands.keys())

        if len(command_list) == 1:
            help_text += f"{command_list[0]} "
        elif len(command_list) <= 4:
            parameters.insert(0, Argument(["command"], type=Choice(command_list)))
        else:
            parameters.insert(0, Argument(["command"]))

    for index, param in enumerate(parameters):
        if param.__class__.__name__ == "Argument":
            if hasattr(param.type, "choices"):
                wrap = "(%s) " if param.required else "[(%s)] "
                help_text += wrap % "|".join(param.type.choices)
            else:
                wrap = "<%s> " if param.required else "[<%s>] "
                help_text += wrap % param.name
        elif param.__class__.__name__ == "Option":
            if (
                parameters[index - 1].__class__.__name__ == "Argument"
                and not help_text.split("\n")[current_line].isspace()
                and len(help_text.split("\n")[current_line]) > 40
            ):
                help_text += "\n" + (" " * indent)
                current_line += 1
            if getattr(param, "is_flag", False):
                wrap = "%s " if param.required else "[%s] "
                options = param.opts
                if getattr(param, "default", None) is None:
                    options += param.secondary_opts
                help_text += wrap % "|".join(options)
            elif hasattr(param.type, "choices"):
                wrap = "%s (%s) " if param.required else "[%s (%s)] "
                help_text += wrap % (param.opts[0], "|".join(param.type.choices))
            else:
                wrap = "%s <%s> " if param.required else "[%s <%s>] "
                help_text += wrap % (param.opts[0], param.name)

        if index + 1 != len(parameters) and len(help_text.split("\n")[current_line]) > 70:
            help_text += "\n" + (" " * indent)
            current_line += 1

    formatter.write(f"{help_text}\n")



from typing import Protocol


class GetVersionsProtocol(Protocol):
    def get_supported_versions(self) -> list[str]: ...


class GetReferenceProtocol(Protocol):
    def get_reference(self) -> str: ...


class AwsGetVersionProtocol(GetReferenceProtocol, GetVersionsProtocol):
    pass
